import json
from textwrap import dedent

from any_agent.evaluation.evaluators import (
    CheckpointEvaluator,
    HypothesisEvaluator,
    QuestionAnsweringSquadEvaluator,
)
from any_agent.evaluation.results_saver import save_evaluation_results
from any_agent.evaluation.test_case import TestCase
from any_agent.logging import logger
from any_agent.telemetry import TelemetryProcessor
from any_agent.tracing import AnyAgentSpan, AnyAgentTrace


def evaluate_telemetry(test_case: TestCase, telemetry_path: str) -> None:
    with open(telemetry_path) as f:
        spans: list[AnyAgentSpan] = json.loads(f.read())
        telemetry = AnyAgentTrace(spans=spans)
    logger.info(f"Telemetry loaded from {telemetry_path}")

    agent_framework = TelemetryProcessor.determine_agent_framework(telemetry)

    processor = TelemetryProcessor.create(agent_framework)
    hypothesis_answer = processor._extract_hypothesis_answer(trace=telemetry)

    checkpoint_evaluator = CheckpointEvaluator(model=test_case.llm_judge)
    checkpoint_results = checkpoint_evaluator.evaluate(
        telemetry=telemetry,
        checkpoints=test_case.checkpoints,
        processor=processor,
    )

    hypothesis_evaluator = HypothesisEvaluator(model=test_case.llm_judge)
    hypothesis_answer_results = hypothesis_evaluator.evaluate(
        hypothesis_final_answer=hypothesis_answer,
        ground_truth_answer_dict=test_case.ground_truth,
        ground_truth_checkpoints=test_case.final_answer_criteria,
    )

    if test_case.ground_truth:
        direct_evaluator = QuestionAnsweringSquadEvaluator()
        direct_results = direct_evaluator.evaluate(
            hypothesis_answer=hypothesis_answer,
            ground_truth_answer=test_case.ground_truth,
        )
    else:
        direct_results = []

    verification_results = (
        checkpoint_results + hypothesis_answer_results + direct_results
    )

    output_message = ""
    output_message += (
        f"""<yellow>Hypothesis Final answer extracted: {hypothesis_answer}</yellow>\n"""
    )
    failed_checks = [r for r in verification_results if not r.passed]
    passed_checks = [r for r in verification_results if r.passed]
    missed_points = sum([r.points for r in failed_checks])
    won_points = sum([r.points for r in passed_checks])
    if passed_checks:
        for check in passed_checks:
            message = dedent(
                f"""
                <green>Passed:
                - {check.criteria}
                - {check.reason}</green>""",
            )
            output_message += message + "\n"
    if failed_checks:
        for check in failed_checks:
            message = dedent(
                f"""
                <red>Failed:
                - {check.criteria}
                - {check.reason}</red>""",
            )
            output_message += message + "\n"
    else:
        output_message += "<green>All checkpoints passed!</green>\n"
    output_message += f"<green>Passed checkpoints: {len(passed_checks)}</green>\n"
    output_message += f"<red>Failed checkpoints: {len(failed_checks)}</red>\n"
    output_message += "<green>=====================================</green>\n"
    output_message += (
        f"<green>Score: {won_points}/{won_points + missed_points}</green>\n"
    )
    output_message += "<green>=====================================</green>\n"
    logger.info(output_message)

    if won_points + missed_points == 0:
        msg = "No points were defined in the test case"
        raise ValueError(msg)
    score = won_points / (won_points + missed_points) * 100

    # Save the evaluation results
    save_evaluation_results(
        test_case=test_case,
        output_path=test_case.output_path,
        output_message=output_message,
        telemetry_path=telemetry_path,
        hypothesis_answer=hypothesis_answer,
        passed_checks=len(passed_checks),
        failed_checks=len(failed_checks),
        score=score,
    )



import os
from abc import ABC, abstractmethod
from contextlib import suppress
from datetime import timedelta
from typing import Any, Literal

from any_agent.config import AgentFramework, MCPSseParams, MCPStdioParams
from any_agent.tools.mcp.mcp_server import MCPServerBase

mcp_available = False
with suppress(ImportError):
    from langchain_mcp_adapters.tools import load_mcp_tools
    from mcp import ClientSession, StdioServerParameters
    from mcp.client.sse import sse_client
    from mcp.client.stdio import stdio_client

    mcp_available = True


class LangchainMCPServerBase(MCPServerBase, ABC):
    client: Any | None = None
    framework: Literal[AgentFramework.LANGCHAIN] = AgentFramework.LANGCHAIN

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,langchain]"
        self.mcp_available = mcp_available
        super()._check_dependencies()

    @abstractmethod
    async def _setup_tools(self) -> None:
        """Set up the LangChain MCP server with the provided configuration."""
        if not self.client:
            msg = "MCP client is not set up. Please call `setup` from a concrete class."
            raise ValueError(msg)

        stdio, write = await self._exit_stack.enter_async_context(self.client)

        client_session = ClientSession(
            stdio,
            write,
            timedelta(seconds=self.mcp_tool.client_session_timeout_seconds)
            if self.mcp_tool.client_session_timeout_seconds
            else None,
        )
        session = await self._exit_stack.enter_async_context(client_session)

        await session.initialize()
        # List available tools
        self.tools = await load_mcp_tools(session)

        self.tools = self._filter_tools(self.tools)


class LangchainMCPServerStdio(LangchainMCPServerBase):
    mcp_tool: MCPStdioParams

    async def _setup_tools(self) -> None:
        server_params = StdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env={**os.environ},
        )

        self.client = stdio_client(server_params)

        await super()._setup_tools()


class LangchainMCPServerSse(LangchainMCPServerBase):
    mcp_tool: MCPSseParams

    async def _setup_tools(self) -> None:
        self.client = sse_client(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
        )

        await super()._setup_tools()


LangchainMCPServer = LangchainMCPServerStdio | LangchainMCPServerSse



import os
from abc import ABC, abstractmethod
from contextlib import suppress
from typing import Literal

from any_agent.config import AgentFramework, MCPSseParams, MCPStdioParams
from any_agent.tools.mcp.mcp_server import MCPServerBase

mcp_available = False
with suppress(ImportError):
    from google.adk.tools.mcp_tool.mcp_toolset import MCPToolset as GoogleMCPToolset
    from google.adk.tools.mcp_tool.mcp_toolset import (  # type: ignore[attr-defined]
        SseServerParams as GoogleSseServerParameters,
    )
    from google.adk.tools.mcp_tool.mcp_toolset import (  # type: ignore[attr-defined]
        StdioServerParameters as GoogleStdioServerParameters,
    )

    mcp_available = True


class GoogleMCPServerBase(MCPServerBase, ABC):
    server: GoogleMCPToolset | None = None
    framework: Literal[AgentFramework.GOOGLE] = AgentFramework.GOOGLE

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,google]"
        self.mcp_available = mcp_available
        super()._check_dependencies()

    @abstractmethod
    async def _setup_tools(self) -> None:
        """Set up the Google MCP server with the provided configuration."""
        if not self.server:
            msg = "MCP server is not set up. Please call `setup` from a concrete class."
            raise ValueError(msg)

        await self._exit_stack.enter_async_context(self.server)
        self.tools = await self.server.load_tools()

        self.tools = self._filter_tools(self.tools)


class GoogleMCPServerStdio(GoogleMCPServerBase):
    mcp_tool: MCPStdioParams

    async def _setup_tools(self) -> None:
        params = GoogleStdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env={**os.environ},
        )
        self.server = GoogleMCPToolset(connection_params=params)

        await super()._setup_tools()


class GoogleMCPServerSse(GoogleMCPServerBase):
    mcp_tool: MCPSseParams

    async def _setup_tools(self) -> None:
        params = GoogleSseServerParameters(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
        )
        self.server = GoogleMCPToolset(connection_params=params)

        await super()._setup_tools()


GoogleMCPServer = GoogleMCPServerStdio | GoogleMCPServerSse



import os
from datetime import datetime
from typing import Any

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.config import MCPStdioParams


def get_current_year() -> str:
    """Get the current year"""
    return str(datetime.now().year)


@pytest.mark.skipif(
    os.environ.get("ANY_AGENT_INTEGRATION_TESTS", "FALSE").upper() != "TRUE",
    reason="Integration tests require `ANY_AGENT_INTEGRATION_TESTS=TRUE` env var",
)
def test_mcp(agent_framework: AgentFramework, tmp_path) -> None:  # type: ignore[no-untyped-def]
    """Get the current year"""
    kwargs: dict[str, Any] = {}

    tools = [
        get_current_year,
        MCPStdioParams(
            command="docker",
            args=[
                "run",
                "-i",
                "--rm",
                "--mount",
                f"type=bind,src={tmp_path},dst=/projects",
                "mcp/filesystem",
                "/projects",
            ],
            tools=[
                "write_file",
            ],
        ),
    ]
    agent_config = AgentConfig(
        model_id="gpt-4.1-mini",
        tools=tools,  # type: ignore[arg-type]
        **kwargs,
    )
    agent = AnyAgent.create(agent_framework, agent_config)
    assert len(agent._mcp_servers) > 0
    result = agent.run(
        "Use the tools to find what year is it and write the value (single number) to /projects/tmp.txt"
    )
    assert os.path.exists(os.path.join(tmp_path, "tmp.txt"))
    with open(os.path.join(tmp_path, "tmp.txt")) as f:
        content = f.read()
    assert content == str(datetime.now().year)
    assert result
    assert result.final_output
    if agent_framework not in [AgentFramework.LLAMA_INDEX]:
        # Llama Index doesn't currently give back raw_responses.
        assert result.raw_responses
        assert len(result.raw_responses) > 0



import os
from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import suppress
from typing import Literal

from any_agent.config import AgentFramework, MCPSseParams, MCPStdioParams
from any_agent.tools.mcp.mcp_server import MCPServerBase

mcp_available = False
with suppress(ImportError):
    from mcp import StdioServerParameters
    from smolagents.mcp_client import MCPClient
    from smolagents.tools import Tool as SmolagentsTool

    mcp_available = True


class SmolagentsMCPServerBase(MCPServerBase, ABC):
    smolagent_tools: Sequence[SmolagentsTool] | None = None
    framework: Literal[AgentFramework.SMOLAGENTS] = AgentFramework.SMOLAGENTS

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,smolagents]"
        self.mcp_available = mcp_available
        super()._check_dependencies()

    @abstractmethod
    async def _setup_tools(self) -> None:
        """Set up the Smolagents MCP server with the provided configuration."""
        if not self.smolagent_tools:
            msg = "Tool collection is not set up. Please call `setup` from a concrete class."
            raise ValueError(msg)

        self.tools = self._filter_tools(self.smolagent_tools)


class SmolagentsMCPServerStdio(SmolagentsMCPServerBase):
    mcp_tool: MCPStdioParams

    async def _setup_tools(self) -> None:
        server_parameters = StdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env={**os.environ},
        )
        self.smolagent_tools = self._exit_stack.enter_context(
            MCPClient(server_parameters)
        )

        await super()._setup_tools()


class SmolagentsMCPServerSse(SmolagentsMCPServerBase):
    mcp_tool: MCPSseParams

    async def _setup_tools(self) -> None:
        server_parameters = {
            "url": self.mcp_tool.url,
        }
        self.smolagent_tools = self._exit_stack.enter_context(
            MCPClient(server_parameters)
        )

        await super()._setup_tools()


SmolagentsMCPServer = SmolagentsMCPServerStdio | SmolagentsMCPServerSse



import os
from abc import ABC, abstractmethod
from contextlib import suppress
from typing import Literal

from any_agent.config import AgentFramework, MCPSseParams, MCPStdioParams
from any_agent.tools.mcp.mcp_server import MCPServerBase

mcp_available = False
with suppress(ImportError):
    from llama_index.tools.mcp import BasicMCPClient as LlamaIndexMCPClient
    from llama_index.tools.mcp import McpToolSpec as LlamaIndexMcpToolSpec

    mcp_available = True


class LlamaIndexMCPServerBase(MCPServerBase, ABC):
    client: LlamaIndexMCPClient | None = None
    framework: Literal[AgentFramework.LLAMA_INDEX] = AgentFramework.LLAMA_INDEX

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,llama_index]"
        self.mcp_available = mcp_available
        super()._check_dependencies()

    @abstractmethod
    async def _setup_tools(self) -> None:
        """Set up the LlamaIndex MCP server with the provided configuration."""
        if not self.client:
            msg = "MCP client is not set up. Please call `setup` from a concrete class."
            raise ValueError(msg)

        mcp_tool_spec = LlamaIndexMcpToolSpec(
            client=self.client,
            allowed_tools=list(self.mcp_tool.tools or []),
        )

        self.tools = await mcp_tool_spec.to_tool_list_async()


class LlamaIndexMCPServerStdio(LlamaIndexMCPServerBase):
    mcp_tool: MCPStdioParams

    async def _setup_tools(self) -> None:
        self.client = LlamaIndexMCPClient(
            command_or_url=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env={**os.environ},
        )

        await super()._setup_tools()


class LlamaIndexMCPServerSse(LlamaIndexMCPServerBase):
    mcp_tool: MCPSseParams

    async def _setup_tools(self) -> None:
        self.client = LlamaIndexMCPClient(command_or_url=self.mcp_tool.url)

        await super()._setup_tools()


LlamaIndexMCPServer = LlamaIndexMCPServerStdio | LlamaIndexMCPServerSse



import pytest
from opentelemetry.sdk.trace import ReadableSpan

from any_agent import AgentFramework, AnyAgentSpan
from any_agent.telemetry import TelemetryProcessor


def test_telemetry_extract_interaction(
    agent_framework: AgentFramework, llm_span: ReadableSpan
) -> None:
    if agent_framework in (
        AgentFramework.AGNO,
        AgentFramework.GOOGLE,
        AgentFramework.TINYAGENT,
    ):
        pytest.skip()
    processor = TelemetryProcessor.create(AgentFramework(agent_framework))
    assert llm_span.attributes  # to make mypy happy
    span = AnyAgentSpan.from_readable_span(llm_span)
    span_kind, interaction = processor.extract_interaction(span)
    assert span_kind == "LLM"
    assert interaction["input"]



from unittest.mock import MagicMock, patch

from opentelemetry.sdk.trace import ReadableSpan

from any_agent.config import AgentFramework, TracingConfig
from any_agent.tracing import RichConsoleSpanExporter


def test_rich_console_span_exporter_default(llm_span: ReadableSpan):  # type: ignore[no-untyped-def]
    console_mock = MagicMock()
    with patch("any_agent.tracing.Console", console_mock):
        exporter = RichConsoleSpanExporter(AgentFramework.LANGCHAIN, TracingConfig())
        exporter.export([llm_span])
        console_mock.return_value.rule.assert_called()


def test_rich_console_span_exporter_disable(llm_span: ReadableSpan):  # type: ignore[no-untyped-def]
    console_mock = MagicMock()
    with patch("any_agent.tracing.Console", console_mock):
        exporter = RichConsoleSpanExporter(
            AgentFramework.LANGCHAIN,
            TracingConfig(llm=None),
        )
        exporter.export([llm_span])
        console_mock.return_value.rule.assert_not_called()


def test_rich_console_cost_info_default(llm_span: ReadableSpan):  # type: ignore[no-untyped-def]
    console_mock = MagicMock()
    with patch("any_agent.tracing.Console", console_mock):
        exporter = RichConsoleSpanExporter(
            AgentFramework.LANGCHAIN,
            TracingConfig(),
        )
        exporter.export([llm_span])
        print_args = [
            args[0][0] for args in console_mock.return_value.print.call_args_list
        ]
        for key in (
            "token_count_prompt",
            "token_count_completion",
            "cost_prompt",
            "cost_completion",
        ):
            assert any(key in arg for arg in print_args)


def test_rich_console_cost_info_disabled(llm_span: ReadableSpan):  # type: ignore[no-untyped-def]
    console_mock = MagicMock()
    with patch("any_agent.tracing.Console", console_mock):
        exporter = RichConsoleSpanExporter(
            AgentFramework.LANGCHAIN,
            TracingConfig(cost_info=False),
        )
        exporter.export([llm_span])
        print_args = [
            args[0][0] for args in console_mock.return_value.print.call_args_list
        ]
        for key in (
            "token_count_prompt",
            "token_count_completion",
            "cost_prompt",
            "cost_completion",
        ):
            assert not any(key in arg for arg in print_args)



# pylint: disable=unused-argument, unused-variable
from collections.abc import Sequence

import pytest

from any_agent.config import AgentFramework, MCPSseParams, MCPStdioParams
from any_agent.tools import _get_mcp_server


@pytest.mark.asyncio
@pytest.mark.usefixtures(
    "_patch_stdio_client",
    "_patch_client_session_initialize",
    "_patch_client_session_list_tools",
)
async def test_stdio_tool_filtering(
    agent_framework: AgentFramework,
    stdio_params: MCPStdioParams,
    tools: Sequence[str],
) -> None:
    server = _get_mcp_server(stdio_params, agent_framework)
    await server._setup_tools()
    if agent_framework == AgentFramework.AGNO:
        # Check that only the specified tools are included
        assert set(server.tools[0].functions.keys()) == set(tools)  # type: ignore[union-attr]
    else:
        assert len(server.tools) == len(tools)  # ignore[arg-type]


@pytest.mark.asyncio
async def test_sse_tool_filtering(
    agent_framework: AgentFramework,
    sse_params_echo_server: MCPSseParams,
    tools: Sequence[str],
) -> None:
    server = _get_mcp_server(sse_params_echo_server, agent_framework)
    await server._setup_tools()
    if agent_framework is AgentFramework.AGNO:
        # Check that only the specified tools are included
        assert set(server.tools[0].functions.keys()) == set(tools)  # type: ignore[union-attr]
    else:
        assert len(server.tools) == len(tools)  # ignore[arg-type]



from collections.abc import Sequence

from any_agent.evaluation.evaluators.LLMEvaluator import LLMEvaluator
from any_agent.evaluation.evaluators.schemas import EvaluationResult
from any_agent.evaluation.test_case import CheckpointCriteria
from any_agent.logging import logger
from any_agent.telemetry import TelemetryProcessor
from any_agent.tracing import AnyAgentTrace


class CheckpointEvaluator(LLMEvaluator):
    """Evaluates checkpoints against telemetry."""

    def evaluate(
        self,
        telemetry: AnyAgentTrace,
        checkpoints: Sequence[CheckpointCriteria],
        processor: TelemetryProcessor,
    ) -> list[EvaluationResult]:
        """Verify each checkpoint against the telemetry data using LLM.

        Args:
            telemetry: The telemetry data to evaluate
            checkpoints: List of checkpoint criteria to verify
            processor: Telemetry processor to extract evidence

        Returns:
            List of evaluation results

        """
        evidence = processor.extract_evidence(telemetry)
        evidence = evidence.replace("<", "\\<").replace(">", "\\>")
        logger.info(f"""<yellow>Evidence\n{evidence}</yellow>\n""")
        results = []

        for checkpoint in checkpoints:
            evaluation = self.llm_evaluate_with_criterion(
                criteria=checkpoint.criteria,
                points=checkpoint.points,
                evidence=evidence,
            )
            results.append(evaluation)

        return results



from .cost import TokenUseAndCost, TotalTokenUseAndCost, extract_token_use_and_cost
from .frameworks.base import TelemetryProcessor

__all__ = [
    "TelemetryProcessor",
    "TokenUseAndCost",
    "TotalTokenUseAndCost",
    "extract_token_use_and_cost",
]



import os
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent, TracingConfig


def test_create_any_with_framework(agent_framework: AgentFramework) -> None:
    agent = AnyAgent.create(agent_framework, AgentConfig(model_id="gpt-4o"))
    assert agent


def test_create_any_with_valid_string(agent_framework: AgentFramework) -> None:
    agent = AnyAgent.create(agent_framework.name, AgentConfig(model_id="gpt-4o"))
    assert agent


def test_create_any_with_invalid_string() -> None:
    with pytest.raises(ValueError, match="Unsupported agent framework"):
        AnyAgent.create("non-existing", AgentConfig(model_id="gpt-4o"))


def test_load_agent_tracing(tmp_path: Path, agent_framework: AgentFramework) -> None:
    mock_agent = MagicMock(spec=AnyAgent)
    mock_agent.load_agent = AsyncMock(return_value=None)

    # Dynamically create the import path based on the framework
    agent_class_path = _get_agent_class_path(agent_framework)

    # Skip frameworks that don't support tracing
    if agent_framework in (
        AgentFramework.AGNO,
        AgentFramework.GOOGLE,
        AgentFramework.TINYAGENT,
    ):
        return

    with patch(agent_class_path, return_value=mock_agent):
        agent = AnyAgent.create(
            agent_framework=agent_framework,
            agent_config=AgentConfig(
                model_id="gpt-4o",
            ),
            tracing=TracingConfig(output_dir=str(tmp_path)),
        )
        # the agent.trace_filepath should be a file that exists
        assert agent.trace_filepath is not None
        assert os.path.exists(agent.trace_filepath)


@pytest.mark.parametrize("framework", list(AgentFramework))
def test_load_agent_no_tracing(framework: AgentFramework) -> None:
    mock_agent = MagicMock(spec=AnyAgent)
    mock_agent.load_agent = AsyncMock(return_value=None)
    mock_agent.trace_filepath = None

    # Dynamically create the import path based on the framework
    agent_class_path = _get_agent_class_path(framework)

    with patch(agent_class_path, return_value=mock_agent):
        agent = AnyAgent.create(
            agent_framework=framework,
            agent_config=AgentConfig(model_id="gpt-4o"),
        )
        # the agent.trace_filepath should be None
        assert agent.trace_filepath is None


def _get_agent_class_path(framework: AgentFramework) -> str:
    """Helper function to get the import path for an agent class based on framework."""
    framework_map = {
        AgentFramework.SMOLAGENTS: "any_agent.frameworks.smolagents.SmolagentsAgent",
        AgentFramework.LANGCHAIN: "any_agent.frameworks.langchain.LangchainAgent",
        AgentFramework.OPENAI: "any_agent.frameworks.openai.OpenAIAgent",
        AgentFramework.LLAMA_INDEX: "any_agent.frameworks.llama_index.LlamaIndexAgent",
        AgentFramework.GOOGLE: "any_agent.frameworks.google.GoogleAgent",
        AgentFramework.AGNO: "any_agent.frameworks.agno.AgnoAgent",
        AgentFramework.TINYAGENT: "any_agent.frameworks.tinyagent.TinyAgent",
    }
    return framework_map[framework]



from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from any_agent.config import AgentFramework, TracingConfig
from any_agent.tracing import Tracer


def test_tracer_initialization(tmp_path: Path) -> None:
    mock_trace = MagicMock()
    mock_tracer_provider = MagicMock()

    with (
        patch("any_agent.tracing.trace", mock_trace),
        patch("any_agent.tracing.TracerProvider", mock_tracer_provider),
    ):
        tracer = Tracer(
            agent_framework=AgentFramework.OPENAI,
            tracing_config=TracingConfig(
                output_dir=str(tmp_path / "traces"),
            ),
        )

        # Verify tracer was initialized correctly
        assert tracer.agent_framework == AgentFramework.OPENAI
        assert tracer.tracing_config.output_dir == str(tmp_path / "traces")
        assert tracer.is_enabled is True
        assert (tmp_path / "traces").exists()

        # Verify tracing was set up
        mock_trace.set_tracer_provider.assert_called_once_with(
            mock_tracer_provider.return_value,
        )


def test_tracer_with_unsupported_framework(tmp_path: Path) -> None:
    with pytest.raises(NotImplementedError, match="AGNO tracing is not supported."):
        Tracer(
            agent_framework=AgentFramework.AGNO,
            tracing_config=TracingConfig(
                output_dir=str(tmp_path / "traces"), console=False
            ),
        )



import os
from abc import ABC, abstractmethod
from contextlib import suppress
from typing import Literal

from any_agent.config import AgentFramework, MCPSseParams, MCPStdioParams
from any_agent.tools.mcp.mcp_server import MCPServerBase

mcp_available = False
with suppress(ImportError):
    from agno.tools.mcp import MCPTools as AgnoMCPTools
    from mcp import ClientSession
    from mcp.client.sse import sse_client

    mcp_available = True


class AgnoMCPServerBase(MCPServerBase, ABC):
    server: AgnoMCPTools | None = None
    framework: Literal[AgentFramework.AGNO] = AgentFramework.AGNO

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,agno]"
        self.mcp_available = mcp_available
        super()._check_dependencies()

    @abstractmethod
    async def _setup_tools(self) -> None:
        """Set up the Agno MCP server with the provided configuration."""
        if not self.server:
            msg = "MCP server is not set up. Please call `setup` from a concrete class."
            raise ValueError(msg)

        self.tools = [await self._exit_stack.enter_async_context(self.server)]  # type: ignore[arg-type]


class AgnoMCPServerStdio(AgnoMCPServerBase):
    mcp_tool: MCPStdioParams

    async def _setup_tools(self) -> None:
        server_params = f"{self.mcp_tool.command} {' '.join(self.mcp_tool.args)}"
        self.server = AgnoMCPTools(
            command=server_params,
            include_tools=list(self.mcp_tool.tools or []),
            env={**os.environ},
        )

        await super()._setup_tools()


class AgnoMCPServerSse(AgnoMCPServerBase):
    mcp_tool: MCPSseParams

    async def _setup_tools(self) -> None:
        client = sse_client(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
        )
        sse_transport = await self._exit_stack.enter_async_context(client)
        stdio, write = sse_transport
        client_session = ClientSession(stdio, write)
        session = await self._exit_stack.enter_async_context(client_session)
        await session.initialize()
        self.server = AgnoMCPTools(
            session=session,
            include_tools=list(self.mcp_tool.tools or []),
        )

        await super()._setup_tools()


AgnoMCPServer = AgnoMCPServerStdio | AgnoMCPServerSse



import os
from pathlib import Path

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent, TracingConfig
from any_agent.tools import search_web


@pytest.mark.skipif(
    os.environ.get("ANY_AGENT_INTEGRATION_TESTS", "FALSE").upper() != "TRUE",
    reason="Integration tests require `ANY_AGENT_INTEGRATION_TESTS=TRUE` env var",
)
def test_load_and_run_agent(agent_framework: AgentFramework, tmp_path: Path) -> None:
    kwargs = {}

    kwargs["model_id"] = "gpt-4.1-nano"
    if "OPENAI_API_KEY" not in os.environ:
        pytest.skip(f"OPENAI_API_KEY needed for {agent_framework}")

    model_args = (
        {"parallel_tool_calls": False}
        if agent_framework is not AgentFramework.AGNO
        else None
    )

    agent_config = AgentConfig(
        tools=[search_web],
        instructions="Search the web to answer",
        model_args=model_args,
        **kwargs,  # type: ignore[arg-type]
    )
    traces = tmp_path / "traces"
    agent = AnyAgent.create(
        agent_framework, agent_config, tracing=TracingConfig(output_dir=str(traces))
    )
    result = agent.run("Which agent framework is the best?")
    assert result
    assert result.final_output
    if agent_framework not in [AgentFramework.LLAMA_INDEX]:
        # Llama Index doesn't currently give back raw_responses.
        assert result.raw_responses
        assert len(result.raw_responses) > 0
    if agent_framework not in (
        AgentFramework.AGNO,
        AgentFramework.GOOGLE,
        AgentFramework.TINYAGENT,
    ):
        assert traces.exists()
        assert agent_framework.name in str(next(traces.iterdir()).name)
        assert result.trace is not None
        assert agent.trace_filepath is not None
        cost_sum = result.trace.get_total_cost()
        assert cost_sum.total_cost > 0
        assert cost_sum.total_cost < 1.00
        assert cost_sum.total_tokens > 0
        assert cost_sum.total_tokens < 20000



from .config import AgentConfig, AgentFramework, TracingConfig
from .frameworks.any_agent import AgentResult, AnyAgent
from .tracing import AnyAgentSpan, AnyAgentTrace

__all__ = [
    "AgentConfig",
    "AgentFramework",
    "AgentResult",
    "AnyAgent",
    "AnyAgentSpan",
    "AnyAgentTrace",
    "TracingConfig",
]



import os
from pathlib import Path

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.config import TracingConfig
from any_agent.tools import search_web, show_final_answer, visit_webpage


@pytest.mark.skipif(
    os.environ.get("ANY_AGENT_INTEGRATION_TESTS", "FALSE").upper() != "TRUE",
    reason="Integration tests require `ANY_AGENT_INTEGRATION_TESTS=TRUE` env var",
)
def test_load_and_run_multi_agent(
    agent_framework: AgentFramework, tmp_path: Path
) -> None:
    kwargs = {}

    if agent_framework is AgentFramework.TINYAGENT:
        pytest.skip(
            f"Skipping test for {agent_framework.name} because it does not support multi-agent"
        )

    kwargs["model_id"] = "gpt-4.1-nano"
    if "OPENAI_API_KEY" not in os.environ:
        pytest.skip(f"OPENAI_API_KEY needed for {agent_framework.name}")

    main_agent = AgentConfig(
        instructions="Use the available agents to complete the task.",
        description="The orchestrator that can use other agents.",
        model_args={"parallel_tool_calls": False}
        if agent_framework is not AgentFramework.AGNO
        else {},
        **kwargs,  # type: ignore[arg-type]
    )

    managed_agents = [
        AgentConfig(
            name="search_web_agent",
            model_id="gpt-4.1-nano",
            description="Agent that can search the web",
            tools=[search_web],
        ),
        AgentConfig(
            name="visit_webpage_agent",
            model_id="gpt-4.1-nano",
            description="Agent that can visit webpages",
            tools=[visit_webpage],
        ),
    ]
    if agent_framework is not AgentFramework.SMOLAGENTS:
        managed_agents.append(
            AgentConfig(
                name="final_answer_agent",
                model_id="gpt-4.1-nano",
                description="Agent that can show the final answer",
                tools=[show_final_answer],
                handoff=True,
            ),
        )
    agent = AnyAgent.create(
        agent_framework=agent_framework,
        agent_config=main_agent,
        managed_agents=managed_agents,
        tracing=TracingConfig(console=False, save=True, cost_info=True),
    )
    result = agent.run("Which agent framework is the best?")

    assert result
    assert result.final_output
    if agent_framework not in [AgentFramework.LLAMA_INDEX]:
        # Llama Index doesn't currently give back raw_responses.
        assert result.raw_responses
        assert len(result.raw_responses) > 0
    if agent_framework not in (
        AgentFramework.AGNO,
        AgentFramework.GOOGLE,
        AgentFramework.TINYAGENT,
    ):
        assert result.trace is not None
        cost_sum = result.trace.get_total_cost()
        assert cost_sum.total_cost > 0
        assert cost_sum.total_cost < 1.00
        assert cost_sum.total_tokens > 0
        assert cost_sum.total_tokens < 20000



from collections.abc import Mapping

from litellm.cost_calculator import cost_per_token
from opentelemetry.util.types import AttributeValue
from pydantic import BaseModel, ConfigDict

from any_agent.logging import logger


class TokenUseAndCost(BaseModel):
    """Token use and cost information."""

    token_count_prompt: int
    token_count_completion: int
    cost_prompt: float
    cost_completion: float

    model_config = ConfigDict(extra="forbid")


class TotalTokenUseAndCost(BaseModel):
    """Total token use and cost information."""

    total_token_count_prompt: int
    total_token_count_completion: int
    total_cost_prompt: float
    total_cost_completion: float

    total_cost: float
    total_tokens: int

    model_config = ConfigDict(extra="forbid")


def extract_token_use_and_cost(
    attributes: Mapping[str, AttributeValue],
) -> TokenUseAndCost:
    """Use litellm and openinference keys to extract token use and cost."""
    span_info: dict[str, AttributeValue] = {}

    for key in ["llm.token_count.prompt", "llm.token_count.completion"]:
        if key in attributes:
            name = key.split(".")[-1]
            span_info[f"token_count_{name}"] = attributes[key]
    try:
        cost_prompt, cost_completion = cost_per_token(
            model=str(attributes.get("llm.model_name", "")),
            prompt_tokens=int(attributes.get("llm.token_count.prompt", 0)),  # type: ignore[arg-type]
            completion_tokens=int(span_info.get("llm.token_count.completion", 0)),  # type: ignore[arg-type]
        )
        span_info["cost_prompt"] = cost_prompt
        span_info["cost_completion"] = cost_completion
    except Exception as e:
        msg = f"Error computing cost_per_token: {e}"
        logger.warning(msg)
        span_info["cost_prompt"] = 0.0
        span_info["cost_completion"] = 0.0

    return TokenUseAndCost.model_validate(span_info)



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, assert_never

from pydantic import BaseModel, ConfigDict

from any_agent.config import AgentConfig, AgentFramework, Tool, TracingConfig
from any_agent.logging import logger
from any_agent.tools.wrappers import _wrap_tools
from any_agent.tracing import AnyAgentTrace, Tracer

if TYPE_CHECKING:
    from collections.abc import Sequence

    from any_agent.tools.mcp.mcp_server import MCPServerBase


class AgentResult(BaseModel):
    """Standardized output format for all agent frameworks."""

    final_output: str | int | float | list[Any] | dict[Any, Any] | None
    raw_responses: list[Any] | None
    trace: AnyAgentTrace | None = None

    model_config = ConfigDict(extra="forbid")


class AnyAgent(ABC):
    """Base abstract class for all agent implementations.

    This provides a unified interface for different agent frameworks.
    """

    def __init__(
        self,
        config: AgentConfig,
        managed_agents: Sequence[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ):
        self.config = config
        self.managed_agents = managed_agents
        self._mcp_servers: list[MCPServerBase] = []
        self._last_tracer: Tracer | None = (
            None  # holds the trace from the most recent run
        )
        self._tracing_config = tracing

    async def _load_tools(
        self, tools: Sequence[Tool]
    ) -> tuple[list[Any], list[MCPServerBase]]:
        tools, mcp_servers = await _wrap_tools(tools, self.framework)
        # Add to agent so that it doesn't get garbage collected
        self._mcp_servers.extend(mcp_servers)
        for mcp_server in mcp_servers:
            tools.extend(mcp_server.tools)
        return tools, mcp_servers

    @staticmethod
    def _get_agent_type_by_framework(
        framework_raw: AgentFramework | str,
    ) -> type[AnyAgent]:
        framework = AgentFramework.from_string(framework_raw)

        if framework is AgentFramework.SMOLAGENTS:
            from any_agent.frameworks.smolagents import SmolagentsAgent

            return SmolagentsAgent

        if framework is AgentFramework.LANGCHAIN:
            from any_agent.frameworks.langchain import LangchainAgent

            return LangchainAgent

        if framework is AgentFramework.OPENAI:
            from any_agent.frameworks.openai import OpenAIAgent

            return OpenAIAgent

        if framework is AgentFramework.LLAMA_INDEX:
            from any_agent.frameworks.llama_index import LlamaIndexAgent

            return LlamaIndexAgent

        if framework is AgentFramework.GOOGLE:
            from any_agent.frameworks.google import GoogleAgent

            return GoogleAgent

        if framework is AgentFramework.AGNO:
            from any_agent.frameworks.agno import AgnoAgent

            return AgnoAgent

        if framework is AgentFramework.TINYAGENT:
            from any_agent.frameworks.tinyagent import TinyAgent

            return TinyAgent

        assert_never(framework)

    @classmethod
    def create(
        cls,
        agent_framework: AgentFramework | str,
        agent_config: AgentConfig,
        managed_agents: list[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ) -> AnyAgent:
        return asyncio.get_event_loop().run_until_complete(
            cls.create_async(
                agent_framework=agent_framework,
                agent_config=agent_config,
                managed_agents=managed_agents,
                tracing=tracing,
            )
        )

    @classmethod
    async def create_async(
        cls,
        agent_framework: AgentFramework | str,
        agent_config: AgentConfig,
        managed_agents: list[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ) -> AnyAgent:
        agent_cls = cls._get_agent_type_by_framework(agent_framework)
        agent = agent_cls(agent_config, managed_agents=managed_agents, tracing=tracing)
        await agent.load_agent()
        return agent

    @property
    def trace_filepath(self) -> str | None:
        """Return the trace filepath."""
        if self._last_tracer is not None:
            return self._last_tracer.trace_filepath
        return None

    def run(self, prompt: str, **kwargs: Any) -> AgentResult:
        """Run the agent with the given prompt."""
        return asyncio.get_event_loop().run_until_complete(
            self.run_async(prompt, **kwargs)
        )

    @abstractmethod
    async def load_agent(self) -> None:
        """Load the agent instance."""

    @abstractmethod
    async def run_async(self, prompt: str, **kwargs: Any) -> AgentResult:
        """Run the agent asynchronously with the given prompt."""

    @property
    @abstractmethod
    def framework(self) -> AgentFramework:
        """The Agent Framework used."""

    @property
    def agent(self) -> Any:
        """The underlying agent implementation from the framework.

        This property is intentionally restricted to maintain framework abstraction
        and prevent direct dependency on specific agent implementations.

        If you need functionality that relies on accessing the underlying agent:
        1. Consider if the functionality can be added to the AnyAgent interface
        2. Submit a GitHub issue describing your use case
        3. Contribute a PR implementing the needed functionality

        Raises:
            NotImplementedError: Always raised when this property is accessed

        """
        msg = "Cannot access the 'agent' property of AnyAgent, if you need to use functionality that relies on the underlying agent framework, please file a Github Issue or we welcome a PR to add the functionality to the AnyAgent class"
        raise NotImplementedError(msg)

    def _get_trace(self) -> AnyAgentTrace | None:
        """Get the trace of the agent."""
        if self._last_tracer is not None:
            return self._last_tracer.get_trace()
        return None

    def _create_tracer(self) -> None:
        """Initialize the tracer for the agent. This is called by each implementation of the agent run_async method."""
        if self._last_tracer is not None:
            self._last_tracer.uninstrument()
        if self._tracing_config is not None:
            # Agno not yet supported https://github.com/Arize-ai/openinference/issues/1302
            # Google ADK not yet supported https://github.com/Arize-ai/openinference/issues/1506
            if self.framework in (
                AgentFramework.AGNO,
                AgentFramework.GOOGLE,
                AgentFramework.TINYAGENT,
            ):
                logger.warning(
                    "Tracing is not yet supported for AGNO and GOOGLE frameworks. "
                )
            else:
                self._last_tracer = Tracer(self.framework, self._tracing_config)



from __future__ import annotations

import json
import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar, assert_never

from any_agent import AgentFramework
from any_agent.logging import logger

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from any_agent import AnyAgentSpan
    from any_agent.tracing import AnyAgentTrace


class TelemetryProcessor(ABC):
    """Base class for processing telemetry data from different agent types."""

    MAX_EVIDENCE_LENGTH: ClassVar[int] = 400

    @classmethod
    def create(cls, agent_framework_raw: AgentFramework | str) -> TelemetryProcessor:
        """Create the appropriate telemetry processor."""
        agent_framework = AgentFramework.from_string(agent_framework_raw)

        if agent_framework is AgentFramework.LANGCHAIN:
            from any_agent.telemetry.frameworks.langchain_telemetry import (
                LangchainTelemetryProcessor,
            )

            return LangchainTelemetryProcessor()
        if agent_framework is AgentFramework.SMOLAGENTS:
            from any_agent.telemetry.frameworks.smolagents_telemetry import (
                SmolagentsTelemetryProcessor,
            )

            return SmolagentsTelemetryProcessor()
        if agent_framework is AgentFramework.OPENAI:
            from any_agent.telemetry.frameworks.openai_telemetry import (
                OpenAITelemetryProcessor,
            )

            return OpenAITelemetryProcessor()
        if agent_framework is AgentFramework.LLAMA_INDEX:
            from any_agent.telemetry.frameworks.llama_index_telemetry import (
                LlamaIndexTelemetryProcessor,
            )

            return LlamaIndexTelemetryProcessor()

        if (
            agent_framework is AgentFramework.GOOGLE
            or agent_framework is AgentFramework.AGNO
            or agent_framework is AgentFramework.TINYAGENT
        ):
            raise NotImplementedError

        assert_never(agent_framework)

    @abstractmethod
    def _extract_hypothesis_answer(self, trace: AnyAgentTrace) -> str:
        """Extract the hypothesis agent final answer from the trace."""

    @abstractmethod
    def _get_agent_framework(self) -> AgentFramework:
        """Get the agent type associated with this processor."""

    @abstractmethod
    def _extract_llm_interaction(self, span: AnyAgentSpan) -> Mapping[str, Any]:
        """Extract interaction details of a span of type LLM."""

    @abstractmethod
    def _extract_tool_interaction(self, span: AnyAgentSpan) -> Mapping[str, Any]:
        """Extract interaction details of a span of type TOOL."""

    @abstractmethod
    def _extract_chain_interaction(self, span: AnyAgentSpan) -> Mapping[str, Any]:
        """Extract interaction details of a span of type CHAIN."""

    @abstractmethod
    def _extract_agent_interaction(self, span: AnyAgentSpan) -> Mapping[str, Any]:
        """Extract interaction details of a span of type AGENT."""

    @staticmethod
    def determine_agent_framework(trace: AnyAgentTrace) -> AgentFramework:
        """Determine the agent type based on the trace.

        These are not really stable ways to find it, because we're waiting on some
        reliable method for determining the agent type. This is a temporary solution.
        """
        for span in trace.spans:
            # This is extremely fragile but there currently isn't
            # any specific key to indicate the agent type
            if span.name == "response":
                logger.info("Agent type is OPENAI")
                return AgentFramework.OPENAI
            if isinstance(
                span.attributes.get("input.value"), str
            ) and "langchain" in str(span.attributes["input.value"]):
                logger.info("Agent type is LANGCHAIN")
                return AgentFramework.LANGCHAIN
            if isinstance(span.attributes.get("smolagents.max_steps"), str):
                logger.info("Agent type is SMOLAGENTS")
                return AgentFramework.SMOLAGENTS
        msg = "Could not determine agent type from trace, or agent type not supported"
        raise ValueError(msg)

    def extract_evidence(self, telemetry: AnyAgentTrace) -> str:
        """Extract relevant telemetry evidence."""
        calls = self._extract_telemetry_data(telemetry)
        return self._format_evidence(calls)

    def _format_evidence(self, calls: Sequence[Mapping[str, Any]]) -> str:
        """Format extracted data into a standardized output format."""
        evidence = f"## {self._get_agent_framework().name} Agent Execution\n\n"

        for idx, call in enumerate(calls, start=1):
            evidence += f"### Call {idx}\n"

            # Truncate any values that are too long
            call = {
                k: (
                    v[: self.MAX_EVIDENCE_LENGTH] + "..."
                    if isinstance(v, str) and len(v) > self.MAX_EVIDENCE_LENGTH
                    else v
                )
                for k, v in call.items()
            }

            # Use ensure_ascii=False to prevent escaping Unicode characters
            evidence += json.dumps(call, indent=2, ensure_ascii=False) + "\n\n"

        return evidence

    @staticmethod
    def parse_generic_key_value_string(text: str) -> dict[str, str]:
        """Parse a string that has items of a dict with key-value pairs separated by '='.

        Only splits on '=' signs, handling quoted strings properly.
        """
        pattern = r"(\w+)=('.*?'|\".*?\"|[^'\"=]*?)(?=\s+\w+=|\s*$)"
        result = {}

        matches = re.findall(pattern, text)
        for key, value in matches:
            # Clean up the key
            key = key.strip()

            # Clean up the value - remove surrounding quotes if present
            if (value.startswith("'") and value.endswith("'")) or (
                value.startswith('"') and value.endswith('"')
            ):
                value = value[1:-1]

            # Store in result dictionary
            result[key] = value

        return result

    def _extract_telemetry_data(
        self,
        telemetry: AnyAgentTrace,
    ) -> list[Mapping[str, Any]]:
        """Extract the agent-specific data from telemetry."""
        calls = []

        for span in telemetry.spans:
            calls.append(self.extract_interaction(span)[1])

        return calls

    def extract_interaction(
        self,
        span: AnyAgentSpan,
    ) -> tuple[str, Mapping[str, Any]]:
        """Extract interaction details from a span."""
        span_kind = span.attributes.get("openinference.span.kind", "")

        if span_kind == "LLM" or "LiteLLMModel.__call__" in span.name:
            return "LLM", self._extract_llm_interaction(span)
        if "tool.name" in span.attributes or span.name.endswith("Tool"):
            return "TOOL", self._extract_tool_interaction(span)
        if span_kind == "CHAIN":
            return "CHAIN", self._extract_chain_interaction(span)
        if span_kind == "AGENT":
            return "AGENT", self._extract_agent_interaction(span)
        logger.warning(f"Unknown span kind: {span_kind}. Span: {span}")
        return "UNKNOWN", {}



"""MCP adapter for Tiny framework."""

import os
from collections.abc import Callable
from contextlib import suppress
from datetime import timedelta
from typing import Any, Literal

from mcp import Tool

from any_agent.config import AgentFramework, MCPSseParams, MCPStdioParams
from any_agent.tools.mcp.mcp_server import MCPServerBase

# Check for MCP dependencies
mcp_available = False
with suppress(ImportError):
    from mcp import ClientSession, StdioServerParameters
    from mcp.client.sse import sse_client
    from mcp.client.stdio import stdio_client

    mcp_available = True


class TinyAgentMCPServerBase(MCPServerBase):
    """MCP adapter for Tiny framework."""

    client: Any | None = None
    framework: Literal[AgentFramework.TINYAGENT] = AgentFramework.TINYAGENT
    libraries: str = "any-agent[mcp]"
    session: Any | None = None

    def _check_dependencies(self) -> None:
        """Check if the required libraries are installed."""
        self.mcp_available = mcp_available
        if not self.mcp_available:
            super()._check_dependencies()

    async def _setup_tools(self) -> None:
        """Set up the MCP tools for TinyAgent."""
        if not self.client:
            msg = "MCP client is not set up. Please call `setup` from a concrete class."
            raise ValueError(msg)

        # Setup the client connection using exit stack to manage lifecycle
        stdio, write = await self._exit_stack.enter_async_context(self.client)

        # Create a client session
        client_session = ClientSession(
            stdio,
            write,
            timedelta(seconds=self.mcp_tool.client_session_timeout_seconds)
            if self.mcp_tool.client_session_timeout_seconds
            else None,
        )

        # Start the session
        self.session: ClientSession = await self._exit_stack.enter_async_context(
            client_session
        )
        if not self.session:
            msg = "Failed to create MCP session"
            raise ValueError(msg)
        await self.session.initialize()

        # Get the available tools from the MCP server using schema
        available_tools = await self.session.list_tools()

        # Filter tools if specific tools were requested
        filtered_tools = self._filter_tools(available_tools.tools)

        # Create callable tool functions
        tool_list = []
        for tool_info in filtered_tools:
            tool_list.append(self._create_tool_from_info(tool_info))

        # Store tools as a list
        self.tools = tool_list

    def _create_tool_from_info(self, tool: Tool) -> Callable[..., Any]:
        """Create a tool function from tool information."""
        tool_name = tool.name
        tool_description = tool.description
        input_schema = tool.inputSchema
        session = self.session
        if not self.session:
            msg = "Not connected to MCP server"
            raise ValueError(msg)

        async def tool_function(*args, **kwargs) -> Any:  # type: ignore[no-untyped-def]
            """Tool function that calls the MCP server."""
            # Combine args and kwargs
            combined_args = {}
            if args and len(args) > 0:
                combined_args = args[0]
            combined_args.update(kwargs)

            if not session:
                msg = "Not connected to MCP server"
                raise ValueError(msg)
            # Call the tool on the MCP server
            try:
                return await session.call_tool(tool_name, combined_args)
            except Exception as e:
                return f"Error calling tool {tool_name}: {e!s}"

        # Set attributes for the tool function
        tool_function.__name__ = tool_name
        tool_function.__doc__ = tool_description
        # this isn't a defined attribute of a callable, but we pass it to tinyagent so that it can use it appropriately
        # when constructing the ToolExecutor.
        tool_function.__input_schema__ = input_schema  # type: ignore[attr-defined]

        return tool_function


class TinyAgentMCPServerStdio(TinyAgentMCPServerBase):
    """MCP adapter for Tiny framework using stdio communication."""

    mcp_tool: MCPStdioParams

    async def _setup_tools(self) -> None:
        server_params = StdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env={**os.environ},
        )

        self.client = stdio_client(server_params)

        await super()._setup_tools()


class TinyAgentMCPServerSse(TinyAgentMCPServerBase):
    """MCP adapter for Tiny framework using SSE communication."""

    mcp_tool: MCPSseParams

    async def _setup_tools(self) -> None:
        self.client = sse_client(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
        )

        await super()._setup_tools()


# Union type for Tiny MCP server implementations
TinyAgentMCPServer = TinyAgentMCPServerStdio | TinyAgentMCPServerSse



from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import AsyncExitStack
from textwrap import dedent
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

from any_agent.config import AgentFramework, MCPParams, Tool


class MCPServerBase(BaseModel, ABC):
    mcp_tool: MCPParams
    framework: AgentFramework
    mcp_available: bool = False
    libraries: str = ""

    _exit_stack: AsyncExitStack = PrivateAttr(default_factory=AsyncExitStack)
    tools: Sequence[Tool] = Field(default_factory=list)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, context: Any) -> None:  # noqa: D102
        self._check_dependencies()

    @abstractmethod
    async def _setup_tools(self) -> None: ...

    @abstractmethod
    def _check_dependencies(self) -> None:
        if self.mcp_available:
            return

        msg = f"You need to `pip install '{self.libraries}'` to use MCP."
        raise ImportError(msg)

    def _filter_tools(self, tools: Sequence[Any]) -> Sequence[Any]:
        # Only add the tools listed in mcp_tool['tools'] if specified
        requested_tools = list(self.mcp_tool.tools or [])
        if not requested_tools:
            return tools
        tools = [tool for tool in tools if tool.name in requested_tools]
        if len(tools) != len(requested_tools):
            tool_names = [tool.name for tool in tools]
            raise ValueError(
                dedent(f"""Could not find all requested tools in the MCP server:
                            Requested: {requested_tools}
                            Set:   {tool_names}"""),
            )
        return tools



from collections.abc import Callable, Mapping, MutableMapping, Sequence
from enum import Enum, auto
from typing import Any, Self

from pydantic import BaseModel, ConfigDict, Field, model_validator


class AgentFramework(str, Enum):
    GOOGLE = auto()
    LANGCHAIN = auto()
    LLAMA_INDEX = auto()
    OPENAI = auto()
    AGNO = auto()
    SMOLAGENTS = auto()
    TINYAGENT = auto()

    @classmethod
    def from_string(cls, value: str | Self) -> Self:
        if isinstance(value, cls):
            return value

        formatted_value = value.strip().upper()
        if formatted_value not in cls.__members__:
            error_message = (
                f"Unsupported agent framework: '{value}'. "
                f"Valid frameworks are: {list(cls.__members__.keys())}"
            )
            raise ValueError(error_message)

        return cls[formatted_value]


class MCPStdioParams(BaseModel):
    command: str
    args: Sequence[str]
    tools: Sequence[str] | None = None
    client_session_timeout_seconds: float | None = 5
    """the read timeout passed to the MCP ClientSession."""

    model_config = ConfigDict(frozen=True, extra="forbid")


class MCPSseParams(BaseModel):
    url: str
    headers: Mapping[str, str] | None = None
    tools: Sequence[str] | None = None
    client_session_timeout_seconds: float | None = 5
    """the read timeout passed to the MCP ClientSession."""

    model_config = ConfigDict(frozen=True)


class TracingConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    console: bool = True
    """Print to console."""

    output_dir: str = "traces"
    """Directory to save traces, if json is enabled"""

    save: bool = True
    """Save to json file."""

    cost_info: bool = True
    """whether json and console logs should include cost information"""

    llm: str | None = "yellow"
    """LLM color in console logs"""

    tool: str | None = "blue"
    """Tool color in console logs"""

    agent: str | None = None
    """Agent color in console logs"""

    chain: str | None = None
    """Chain color in console logs"""

    @model_validator(mode="after")
    def validate_enable_flags(self) -> Self:
        if not self.console and not self.save:
            msg = "At least one of `console` or `save` must be true"
            raise ValueError(msg)
        return self

    @model_validator(mode="after")
    def validate_output_dir(self) -> Self:
        if self.save and not self.output_dir:
            msg = "output_dir must be set if json is enabled"
            raise ValueError(msg)
        return self


MCPParams = MCPStdioParams | MCPSseParams

Tool = str | MCPParams | Callable[..., Any]


class AgentConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")
    model_id: str
    api_base: str | None = None
    api_key: str | None = None
    description: str | None = None
    name: str = "any_agent"
    instructions: str | None = None
    tools: Sequence[Tool] = Field(default_factory=list)
    handoff: bool = False
    agent_type: Callable[..., Any] | None = None
    agent_args: MutableMapping[str, Any] | None = None
    model_type: Callable[..., Any] | None = None
    model_args: MutableMapping[str, Any] | None = None



import json
import os
from collections.abc import Mapping, Sequence
from datetime import datetime
from typing import Any, Protocol, assert_never

from opentelemetry import trace
from opentelemetry.sdk.trace import ReadableSpan, TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    SimpleSpanProcessor,
    SpanExporter,
    SpanExportResult,
)
from pydantic import BaseModel, ConfigDict
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel

from any_agent.config import AgentFramework, TracingConfig
from any_agent.logging import logger
from any_agent.telemetry import (
    TelemetryProcessor,
    TokenUseAndCost,
    TotalTokenUseAndCost,
    extract_token_use_and_cost,
)
from any_agent.telemetry.types import (
    AttributeValue,
    Link,
    SpanContext,
    SpanKind,
    Status,
)
from any_agent.telemetry.types import (
    Event as EventModel,
)
from any_agent.telemetry.types import (
    Resource as ResourceModel,
)


# only keep a few things that we care about from the AnyAgentSpan,
# but move it to this class because otherwise we can't recreate it
class AnyAgentSpan(BaseModel):
    """A span that can be exported to JSON or printed to the console."""

    name: str
    kind: SpanKind
    parent: SpanContext | None = None
    start_time: int | None = None
    end_time: int | None = None
    status: Status
    context: SpanContext
    attributes: dict[str, Any]
    links: list[Link]
    events: list[EventModel]
    resource: ResourceModel

    model_config = ConfigDict(arbitrary_types_allowed=False)

    @classmethod
    def from_readable_span(cls, readable_span: ReadableSpan) -> "AnyAgentSpan":
        """Create an AnyAgentSpan from a ReadableSpan."""
        return cls(
            name=readable_span.name,
            kind=SpanKind.from_otel(readable_span.kind),
            parent=SpanContext.from_otel(readable_span.parent),
            start_time=readable_span.start_time,
            end_time=readable_span.end_time,
            status=Status.from_otel(readable_span.status),
            context=SpanContext.from_otel(readable_span.context),
            attributes=dict(readable_span.attributes)
            if readable_span.attributes
            else {},
            links=[Link.from_otel(link) for link in readable_span.links],
            events=[EventModel.from_otel(event) for event in readable_span.events],
            resource=ResourceModel.from_otel(readable_span.resource),
        )

    def set_attributes(self, attributes: Mapping[str, AttributeValue]) -> None:
        """Set attributes for the span."""
        for key, value in attributes.items():
            if key in self.attributes:
                logger.warning("Overwriting attribute %s with %s", key, value)
            self.attributes[key] = value

    def set_attribute(self, key: str, value: AttributeValue) -> None:
        """Set a single attribute for the span."""
        return self.set_attributes({key: value})


class AnyAgentTrace(BaseModel):
    """A trace that can be exported to JSON or printed to the console."""

    spans: list[AnyAgentSpan]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def get_total_cost(self) -> TotalTokenUseAndCost:
        """Return the current total cost and token usage statistics."""
        costs: list[TokenUseAndCost] = []
        for span in self.spans:
            if span.attributes and "cost_prompt" in span.attributes:
                token_use_and_cost = TokenUseAndCost(
                    token_count_prompt=span.attributes["llm.token_count.prompt"],
                    token_count_completion=span.attributes[
                        "llm.token_count.completion"
                    ],
                    cost_prompt=span.attributes["cost_prompt"],
                    cost_completion=span.attributes["cost_completion"],
                )
                costs.append(token_use_and_cost)

        total_cost = sum(cost.cost_prompt + cost.cost_completion for cost in costs)
        total_tokens = sum(
            cost.token_count_prompt + cost.token_count_completion for cost in costs
        )
        total_token_count_prompt = sum(cost.token_count_prompt for cost in costs)
        total_token_count_completion = sum(
            cost.token_count_completion for cost in costs
        )
        total_cost_prompt = sum(cost.cost_prompt for cost in costs)
        total_cost_completion = sum(cost.cost_completion for cost in costs)
        return TotalTokenUseAndCost(
            total_cost=total_cost,
            total_tokens=total_tokens,
            total_token_count_prompt=total_token_count_prompt,
            total_token_count_completion=total_token_count_completion,
            total_cost_prompt=total_cost_prompt,
            total_cost_completion=total_cost_completion,
        )


class TraceSpanExporter(SpanExporter):  # noqa: D101
    def __init__(
        self,
        agent_framework: AgentFramework,
        tracing_config: TracingConfig,
        file_name: str,
        tracer_trace: AnyAgentTrace,
    ):
        """Initialize the JsonFileSpanExporter."""
        self.tracer_trace: AnyAgentTrace = tracer_trace  # so that this exporter can set the trace and can be used to get the trace
        self.processor = TelemetryProcessor.create(agent_framework)
        self.file_name = file_name
        self.tracing_config = tracing_config
        self.save: bool = tracing_config.save
        if self.save and not os.path.exists(self.tracing_config.output_dir):
            os.makedirs(self.tracing_config.output_dir)
        if self.save and not os.path.exists(self.file_name):
            with open(self.file_name, "w", encoding="utf-8") as f:
                json.dump([], f)

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:  # noqa: D102
        # Add new spans
        new_spans = []
        for readable_span in spans:
            span = AnyAgentSpan.from_readable_span(readable_span)
            try:
                # Try to parse the span data from to_json() if it returns a string
                span_kind, _ = self.processor.extract_interaction(span)
                if span_kind == "LLM" and self.tracing_config.cost_info:
                    cost_info = extract_token_use_and_cost(span.attributes)
                    span.set_attributes(cost_info.model_dump())

            except (json.JSONDecodeError, TypeError, AttributeError) as e:
                logger.warning("Failed to parse span data, %s, %s", span, e)
                continue
            new_spans.append(span)
        self.tracer_trace.spans.extend(new_spans)
        with open(self.file_name, "w", encoding="utf-8") as f:
            json.dump(
                [span.model_dump_json() for span in self.tracer_trace.spans],
                f,
                indent=2,
            )

        return SpanExportResult.SUCCESS


class RichConsoleSpanExporter(SpanExporter):  # noqa: D101
    def __init__(self, agent_framework: AgentFramework, tracing_config: TracingConfig):  # noqa: D107
        self.processor = TelemetryProcessor.create(agent_framework)
        self.console = Console()
        self.tracing_config = tracing_config

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:  # noqa: D102
        for readable_span in spans:
            style = None
            span = AnyAgentSpan.from_readable_span(readable_span)
            try:
                span_kind, interaction = self.processor.extract_interaction(span)

                style = getattr(self.tracing_config, span_kind.lower(), None)

                if not style or interaction == {}:
                    continue

                self.console.rule(
                    span_kind,
                    style=style,
                )
                for key, value in interaction.items():
                    if key == "output":
                        self.console.print(
                            Panel(
                                Markdown(str(value or "")),
                                title="Output",
                            ),
                        )
                    else:
                        self.console.print(f"{key}: {value}")

                if span_kind == "LLM" and self.tracing_config.cost_info:
                    cost_info = extract_token_use_and_cost(span.attributes)
                    for key, value in cost_info.model_dump().items():
                        self.console.print(f"{key}: {value}")

            except Exception:
                self.console.print_exception()
            if style:
                self.console.rule(style=style)
        return SpanExportResult.SUCCESS


class Instrumenter(Protocol):  # noqa: D101
    def instrument(self, *, tracer_provider: TracerProvider) -> None: ...  # noqa: D102

    def uninstrument(self) -> None: ...  # noqa: D102


def _get_instrumenter_by_framework(framework: AgentFramework) -> Instrumenter:
    if framework is AgentFramework.OPENAI:
        from openinference.instrumentation.openai_agents import OpenAIAgentsInstrumentor

        return OpenAIAgentsInstrumentor()

    if framework is AgentFramework.SMOLAGENTS:
        from openinference.instrumentation.smolagents import SmolagentsInstrumentor

        return SmolagentsInstrumentor()

    if framework is AgentFramework.LANGCHAIN:
        from openinference.instrumentation.langchain import LangChainInstrumentor

        return LangChainInstrumentor()

    if framework is AgentFramework.LLAMA_INDEX:
        from openinference.instrumentation.llama_index import LlamaIndexInstrumentor

        return LlamaIndexInstrumentor()

    if (
        framework is AgentFramework.GOOGLE
        or framework is AgentFramework.AGNO
        or framework is AgentFramework.TINYAGENT
    ):
        msg = f"{framework} tracing is not supported."
        raise NotImplementedError(msg)

    assert_never(framework)


class Tracer:
    """Tracer is responsible for managing all things tracing for an agent."""

    def __init__(
        self,
        agent_framework: AgentFramework,
        tracing_config: TracingConfig,
    ):
        """Initialize the Tracer and set up tracing filepath, if enabled."""
        self.agent_framework = agent_framework
        # Set it to None at first To avoid AttributeError on __del__ if _get_instrumenter_by_framework throws exception
        self.instrumenter: Instrumenter | None = None
        self.instrumenter = _get_instrumenter_by_framework(
            agent_framework
        )  # Fail fast if framework is not supported
        self.tracing_config = tracing_config
        self.trace_filepath: str | None = None
        self.trace: AnyAgentTrace = AnyAgentTrace(spans=[])

        self._setup_tracing()

    def uninstrument(self) -> None:
        """Uninstrument the tracer."""
        if self.instrumenter:
            self.instrumenter.uninstrument()
            self.instrumenter = None

    def __del__(self) -> None:
        """Stop the openinference instrumentation when the tracer is deleted."""
        self.uninstrument()

    def _setup_tracing(self) -> None:
        """Set up tracing for the agent."""
        tracer_provider = TracerProvider()

        if not self.instrumenter:
            msg = "Instrumenter not found for the agent framework."
            raise ValueError(msg)

        timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        self.trace_filepath = f"{self.tracing_config.output_dir}/{self.agent_framework.name}-{timestamp}.json"
        json_file_exporter = TraceSpanExporter(
            agent_framework=self.agent_framework,
            tracing_config=self.tracing_config,
            file_name=self.trace_filepath,
            tracer_trace=self.trace,
        )
        span_processor = SimpleSpanProcessor(json_file_exporter)
        tracer_provider.add_span_processor(span_processor)

        if self.tracing_config.console:
            processor = BatchSpanProcessor(
                RichConsoleSpanExporter(self.agent_framework, self.tracing_config),
            )
            tracer_provider.add_span_processor(processor)

        trace.set_tracer_provider(tracer_provider)

        self.instrumenter.instrument(tracer_provider=tracer_provider)

    @property
    def is_enabled(self) -> bool:
        """Whether tracing is enabled."""
        return self.tracing_config.save or self.tracing_config.console

    def get_trace(self) -> AnyAgentTrace | None:
        """Return the trace data if file tracing is enabled."""
        if self.trace:
            return self.trace
        msg = "Tracing is not enabled."
        logger.warning(msg)
        return None



from abc import ABC, abstractmethod
from contextlib import suppress
from typing import Literal

from any_agent.config import AgentFramework, MCPSseParams, MCPStdioParams
from any_agent.tools.mcp.mcp_server import MCPServerBase

mcp_available = False
with suppress(ImportError):
    from agents.mcp import MCPServerSse as OpenAIInternalMCPServerSse
    from agents.mcp import (
        MCPServerSseParams as OpenAIInternalMCPServerSseParams,
    )
    from agents.mcp import MCPServerStdio as OpenAIInternalMCPServerStdio
    from agents.mcp import (
        MCPServerStdioParams as OpenAIInternalMCPServerStdioParams,
    )

    mcp_available = True


class OpenAIMCPServerBase(MCPServerBase, ABC):
    server: OpenAIInternalMCPServerStdio | OpenAIInternalMCPServerSse | None = None
    framework: Literal[AgentFramework.OPENAI] = AgentFramework.OPENAI

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,openai]"
        self.mcp_available = mcp_available
        super()._check_dependencies()

    @abstractmethod
    async def _setup_tools(self) -> None:
        """Set up the OpenAI MCP server with the provided configuration."""
        if not self.server:
            msg = "MCP server is not set up. Please call `setup` from a concrete class."
            raise ValueError(msg)

        await self._exit_stack.enter_async_context(self.server)
        self.tools = await self.server.list_tools()  # type: ignore[assignment]

        self.tools = self._filter_tools(self.tools)


class OpenAIMCPServerStdio(OpenAIMCPServerBase):
    mcp_tool: MCPStdioParams

    async def _setup_tools(self) -> None:
        params = OpenAIInternalMCPServerStdioParams(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
        )

        self.server = OpenAIInternalMCPServerStdio(
            name="OpenAI MCP Server",
            params=params,
        )

        await super()._setup_tools()


class OpenAIMCPServerSse(OpenAIMCPServerBase):
    mcp_tool: MCPSseParams

    async def _setup_tools(self) -> None:
        params = OpenAIInternalMCPServerSseParams(url=self.mcp_tool.url)

        self.server = OpenAIInternalMCPServerSse(
            name="OpenAI MCP Server", params=params
        )

        await super()._setup_tools()


OpenAIMCPServer = OpenAIMCPServerStdio | OpenAIMCPServerSse
